{ Install this component using Options|Install Compenents.
  The function of this module is to provide Delphi 1.x with a
  component capable of performing multiple row selection on a DBGrid.

  The code herein is released to the public domain under the condition
  that it will not be used for commercial or "For Profit" ventures.
  This code is considered swap-ware.  If you use it, hack around with it etc.
  Yu must send me something in return.  Weather it be a postcard (minimum) or
  some cash (highly recomemended if you use it in a comercial program) or whatever.

  Written By:      Neal Walfield
  Date:            July 16th, 1997.
  Copyright:       (R) Copyright by Neal Walfield, 1997. All Rights Reserved
  UserID(s):       brin@star.net
                   Contentio@geocities.com  (More Stable)
  Address:         4 Wiseman Dr.
                   Peabody, MA 01960
                   USA

  Description:     This control allows for multiple selection in a DBGrid.

  TMultiGrid v1.00
}
unit Multgrid;
{Note :
 This component maintains, at a maximum Half A Million Rows.  Anymore,
  and it won't maintain it properly.
 Currently, this component doesn't allow multiple selection of cells.
   ie This component is ALWAYS in dgRowSelect mode.  This shouldn't
   be to hard to implement, but, I didn't need it, so I didn't write the
   routines to handle it.  Sorry.
}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Grids, DBGrids, DB, StdCtrls, DBTables, DbiProcs,
  DbiTypes, DbiErrs;

type
  EIndexOutOfBounds = class(Exception); {Occurs when Accessing a non-existtent record}

  Select = (dbSelected, dbUnSelected);

  TMultiGrid = class(TDBGrid)
  private
    { Private declarations }
    Sel           : PChar;                        {Contains what rows are selected}
                    {Dynamically resizes for most effective memory usage.  Can    }
                    {contain a Grid that has 65536*8 Rows or 524,288 rows.        }
    LastPressed   : LongInt;                      {Contains the last pressed row  }
                    {For use internally, (when user presses control or shift etc. }
    function GetSelected(Index : longint) : Select; {Returns the value of a row   }
    procedure SetSelected(Index : longint; S : Select); {Sets a row               }
    procedure CheckIncrease;  {Dynamic Memory Allocation, Called automatically    }
  protected
    { Protected declarations }
    procedure KeyDown(var Key: Word; Shift: TShiftState); override; {Handles key presses}
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override; {Handles mouse clicks.                            }
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DrawDataCell (const Rect: TRect; Field: TField;
      State: TGridDrawState); override; {Handles drawing of grid                            }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override; {Called automatically,      }
                    {creates some memory, and sets default values                 }
    destructor Destroy; override; {Called when TMultiGrid is destroyed. Destroys  }
                    {Memory, and does other housework.                            }
    procedure Clear;  {Reset the Sel var}
    function GetRecordNumber : Longint; {Returns the currently selected record    }
                    {number pointed to by the indiquator.                         }
    property Selected[Index : longint] : Select read GetSelected write SetSelected;
                    {Creates the illusion of an array of values that can be set   }
                    {and deselected.  ie:                                         }
                    {   TMultiGrid.Selected[i] := dbSelected;                     }
                    {   if TMultiGrid.Selected[i] = dnUnSelected then ...         }
    function CurrentIsSelected : Boolean; {Answers quesiton, is the record pointed}
                    { to by the indiquator selected.  Equivalent to:              }
                    {    not Boolean(TMultiGrid.Selected[GetRecordNumber]         }

    function LastSelected: LongInt;
  published
    { Published declarations }
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Data Controls', [TMultiGrid]);
end;

{-------------------------------------------------------------------}
{Dynamic Memory allocation.                                         }
{-------------------------------------------------------------------}

constructor TMultiGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  Sel            := StrAlloc(128); {Start with 512bytes}
  Clear;
end;

destructor TMultiGrid.Destroy;
begin
  StrDispose(Sel);
  inherited Destroy;
end;

{Controls dynamic allocation of memory for the seleced variable.    }
{Allocates memory in bursts of 128 bytes, for greater effectiveness.}
procedure TMultiGrid.CheckIncrease;
var
  Temp : PChar;
  var i : longint;
begin
  {Alloc memory until we have enough, to hold all}
  if DataSource.DataSet.RecordCount shl 3 >= StrBufSize(Sel) then begin
    repeat
      Temp := StrAlloc(StrBufSize(Sel) + 128);
      for i := StrBufSize(Sel) to (StrBufSize(Temp)-1) do
        Temp[i] := #255; {unselected}
      StrCopy(Temp, Sel);
      StrDispose(Sel);
      Sel := Temp;
    until DataSource.DataSet.RecordCount shl 3 < StrBufSize(Sel); {Alloc as much mem as needed plus a little extra}
  end;
 {Dealloc some memory, Only 128bytes at a time, will
   leave some room and save some time}
  if (DataSource.DataSet.RecordCount + 128) shl 3 < StrBufSize(Sel) then begin
    Temp := StrAlloc(StrBufSize(Sel)-128);
    StrLCopy(Temp, Sel, StrBufSize(Sel));
    StrDispose(Sel);
    Sel := Temp;
  end;
end;

{---------------------------------------------------------------------}
{User information Routines, Checking and setting of Row Selection     }
{---------------------------------------------------------------------}

{Reruns the currently selected record.                                }
{ ie The one that the indiquator is pointing to.                      }
function TMultiGrid.GetRecordNumber : Longint;
var
  CursorProps: CurProps;
  RecordProps: RECProps;
begin
  { Return 0 if dataset is not Paradox or dBASE }
  Result := 0;
  with DataSource.DataSet do
  begin
    { Is the dataset active? }
    if State = dsInactive then
      exit;

    {Is table empty?}
    if RecordCount = 0 then exit;

    { We need to make this call to grab the cursor's iSeqNums }
    Check(DbiGetCursorProps(Handle, CursorProps));

    { Synchronize the BDE cursor with the Dataset's cursor }
    UpdateCursorPos;

    { Fill RecordProps with the current record's properties }
    try
      Check(DbiGetRecord(Handle, dbiNOLOCK, nil, @RecordProps));
    finally
    end;  

    { What kind of dataset are we looking at? }
    case CursorProps.iSeqNums of
      0: Result := RecordProps.iPhyRecNum;  { dBASE   }
      1: Result := RecordProps.iSeqNum;     { Paradox }
    end;
  end;
end;

{This routine returns the last selected record, very useful for
 deleting all the selected record etc. (ie When you delete, the
 selected bars are still in the same place, so you must deselect
 after each deletion.  If you started from fron, then you would
 delete a record and all the record swould be off one)
}
function TMultiGrid.LastSelected: LongInt;
var i : longint;
begin
  for i := DataSource.DataSet.RecordCount downto 0 do
    if Selected[i] = dbSelected then begin
      Result := i;
      exit;
    end;
  Result := -1;
end;

{Returns if Record #Index is selected.  This routine is never used    }
{directly. Used as: If TMultiGrid.Selected[i] = dnSelected            }
function TMultiGrid.GetSelected(Index : longint) : Select;
begin
  if Index > StrBufSize(Sel) shl 3 then begin
    CheckIncrease; {Out of bounds, try allocing memory}
    if Index > StrBufSize(Sel) shl 3 then begin{Out of bounds}
      exit;
    end;
  end;

  Result := Select((Integer(Sel[Index div 8]) shr ((Index mod 8))) and 1);
end;

{Sets Record #Index to S (dbSelected/dbUnSelected) This routine is    }
{never called directly.  Always through:                              }
{  TMultiGrid.Selected[i] := dbSelected                               }
procedure TMultiGrid.SetSelected(Index : longint; S : Select);
var
  Temp : LongInt;
begin
  Temp := StrBufSize(Sel);
  if Index shr 3  > StrBufSize(Sel) then begin
    CheckIncrease; {Out of bounds, try allocing memory}
    if Index shr 3 > StrBufSize(Sel) then begin{Out of bounds}
      exit;
    end;
  end;

  if S = dbunSelected then
    Sel[Index div 8] := Char(Integer(Sel[Index div 8]) or (1 shl (Index mod 8)))
  else
    Sel[Index div 8] := Char(Integer(Sel[Index div 8]) and not(1 shl (Index mod 8)))
end;

{Answers the question, is the Record pointed to by the Indiquator     }
{selected?                                                            }
function TMultiGrid.CurrentIsSelected : Boolean;
begin
  result := not Boolean(Selected[getRecordNumber]);
end;

{Sets all the records in teh Grid to dbunSelected                     }
procedure TMultiGrid.Clear;
var
  i : LongInt;
begin
  for i := 0 to (StrBufSize(Sel)-1) shl 3 do
    Selected[i] := dbUnselected;
end;

{---------------------------------------------------------------------}
{The following routines control the selection and delselection of     }
{records when Keys are Pressed or The mouse is pressed.               }
{Selection and deselection works in the same way as a TListBox        }
{---------------------------------------------------------------------}

{Reacts when the user presses a key (ie Up or Down)                   }
procedure TMultiGrid.KeyDown(var Key: Word; Shift: TShiftState);
var
  PhysicalRecord : LongInt;
  Spot : TPoint;
begin
  inherited KeyDown(Key, Shift);

  Spot := BoxRect(1, Row, 1, Row).TopLeft;

  if (Key = 40) {Down} then
    MouseDown(mbLeft, Shift, Spot.X+1, Spot.Y+1);
  if (Key = 38) {Up} then
    MouseDown(mbLeft, Shift, Spot.X+1, Spot.Y+1);
end;

{Reacts when the user clicks the mouse.                               }
procedure TMultiGrid.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  i : LongInt;
  Spot : TGridCoord;
  PhysicalRecord : LongInt;
begin
  CheckIncrease;

  inherited MouseDown(Button, Shift, X, Y);

  if (X >= Width) or (X < 0) or (y < 0) or (Y >= Height) then Exit;

  PhysicalRecord := GetRecordNumber - Row;

  if (ssShift in Shift) and (Button = mbLeft) then begin {Range Select}
    if PhysicalRecord + Row < LastPressed then
      for i := PhysicalRecord + Row to LastPressed do
        Selected[i] := dbSelected
    else
      for i := PhysicalRecord + Row downto LastPressed do
        Selected[i] := dbSelected;
    LastPressed := PhysicalRecord + Row;
    Refresh;
    exit;
  end;

  LastPressed := GetRecordNumber;
  Spot := MouseCoord(X, Y);
  Row := Spot.Y;

  if (ssCtrl in Shift) and (Button = mbLeft) then begin {Add}
    if Selected[PhysicalRecord + Row] = dbSelected then
      Selected[PhysicalRecord + Row] := dbUnSelected
    else
      Selected[PhysicalRecord + Row] := dbSelected;
    LastPressed := PhysicalRecord + Row;
    Refresh;
    exit;
  end;

  if (Button = mbLeft) then begin  {Click}
    Clear;
    Selected[PhysicalRecord + Row] := dbselected;
    LastPressed := PhysicalRecord + Row;
    Refresh;
    exit;
  end;

  if ssDouble in Shift then
    OnDblClick(Self);
end;

procedure TMultiGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Spot : TGridCoord;
  PhysicalRecord : LongInt;
begin
  inherited MouseMove(Shift, X, Y);

  PhysicalRecord := GetRecordNumber - Row;

  { Drag Mouse }
  if (ssLeft in Shift) then begin
    Spot := MouseCoord(X, Y);
    Selected[PhysicalRecord + Spot.Y] := dbselected;
    LastPressed := PhysicalRecord + Spot.Y;
    Refresh;
    exit;
  end;
end;

{---------------------------------------------------------------------}
{The following procedure does all the drawing of the cells            }
{---------------------------------------------------------------------}
procedure TMultiGrid.DrawDataCell(const Rect: TRect; Field: TField;
  State: TGridDrawState);
var
  Spot : TGridCoord;
  PhysicalRecord : LongInt;
begin
  inherited DrawDataCell(Rect, Field, State);

  Spot := MouseCoord(Rect.Left+1, Rect.Top+1);
  PhysicalRecord := GetRecordNumber;

  if Selected[PhysicalRecord] = dbSelected then
    Canvas.Brush.Color := clActiveCaption
  else
    Canvas.Brush.Color := clInActiveCaption;

  Canvas.FillRect(Rect);
  Canvas.TextRect(Rect,Rect.Left,Rect.Top,Field.Text);
end;

end.
